use std::borrow::BorrowMut;

use godot::{engine::Skeleton3D, prelude::*};

use crate::{
    common::{map_blend_shape, map_range_clamped, ReadBlendShape},
    face_data_object::FaceDataObject,
    face_transform_rotation::TransformAxis,
    live_link_server::LiveLinkServer,
};

#[derive(GodotClass)]
#[class(base=Node)]
pub struct FaceBone {
    #[export]
    read_from: Option<Gd<LiveLinkServer>>,
    #[export]
    read_shape: ReadBlendShape,

    #[export]
    target_skeleton: Option<Gd<Skeleton3D>>,
    #[export]
    target_bone: GString,

    #[export]
    clamp_in_min: f32,
    #[export]
    clamp_in_max: f32,
    #[export]
    clamp_out_min: f32,
    #[export]
    clamp_out_max: f32,

    #[export]
    transform_axis: TransformAxis,

    pre_value: f32,

    base: Base<Node>,
}

#[godot_api]
impl FaceBone {
    #[func]
    fn on_face_data(&mut self, changed: bool, face_data: Option<Gd<FaceDataObject>>) {
        if changed {
            let face_data = face_data.unwrap();
            let shape_val = map_blend_shape(&self.read_shape, face_data);
            self.pre_value = shape_val;
        }
    }
}

#[godot_api]
impl INode for FaceBone {
    fn init(base: Base<Node>) -> Self {
        Self {
            read_from: None,
            read_shape: ReadBlendShape::EyeBlinkLeft,
            target_skeleton: None,
            target_bone: GString::default(),
            clamp_in_min: 0.0,
            clamp_in_max: 1.0,
            clamp_out_min: 0.0,
            clamp_out_max: 1.0,

            transform_axis: TransformAxis::X,

            pre_value: 0.0,

            base,
        }
    }

    fn ready(&mut self) {
        if let Some(mut server) = self.borrow_mut().read_from.take() {
            let callable = self.base_mut().callable("on_face_data");
            server.connect("face_data_updated".into(), callable);
        }
    }

    fn process(&mut self, _delta: f64) {
        let skeleton = self.get_target_skeleton();
        if let Some(mut target) = skeleton {
            let bone_id = target.find_bone(self.get_target_bone());
            let initial_rotation = target.get_bone_pose_rotation(bone_id);
            let mut rot_euler = initial_rotation.to_euler(EulerOrder::YXZ);
            let clamped = map_range_clamped(
                self.pre_value,
                self.clamp_in_min,
                self.clamp_in_max,
                self.clamp_out_min,
                self.clamp_out_max,
            );
            match self.transform_axis {
                TransformAxis::X => {
                    rot_euler.x = clamped;
                }
                TransformAxis::Y => {
                    rot_euler.y = clamped;
                }
                TransformAxis::Z => {
                    rot_euler.z = clamped;
                }
            }
            target.set_bone_pose_rotation(bone_id, Quaternion::from_euler(rot_euler).normalized());
        }
    }
}
