use std::borrow::BorrowMut;

use godot::prelude::*;

use crate::{
    common::{map_blend_shape, map_range_clamped, ReadBlendShape},
    face_data_object::FaceDataObject,
    live_link_server::LiveLinkServer,
};

#[derive(GodotClass)]
#[class(base=Node)]
pub struct FaceProperty {
    #[export]
    read_from: Option<Gd<LiveLinkServer>>,
    #[export]
    read_shape: ReadBlendShape,

    #[export]
    target_object: Option<Gd<Node>>,
    #[export]
    target_property: StringName,

    #[export]
    clamp_in_min: f32,
    #[export]
    clamp_in_max: f32,
    #[export]
    clamp_out_min: f32,
    #[export]
    clamp_out_max: f32,

    pre_value: f32,

    base: Base<Node>,
}

#[godot_api]
impl FaceProperty {
    #[func]
    fn on_face_data(&mut self, changed: bool, face_data: Option<Gd<FaceDataObject>>) {
        if changed {
            let face_data = face_data.unwrap();
            let shape_val = map_blend_shape(&self.read_shape, face_data);
            self.pre_value = shape_val;
        }
    }
}

#[godot_api]
impl INode for FaceProperty {
    fn init(base: Base<Node>) -> Self {
        Self {
            read_from: None,
            read_shape: ReadBlendShape::EyeBlinkLeft,
            target_object: None,
            target_property: StringName::default(),
            clamp_in_min: 0.0,
            clamp_in_max: 1.0,
            clamp_out_min: 0.0,
            clamp_out_max: 1.0,

            pre_value: 0.0,

            base,
        }
    }

    fn ready(&mut self) {
        if let Some(mut server) = self.borrow_mut().read_from.take() {
            let callable = self.base_mut().callable("on_face_data");
            server.connect("face_data_updated".into(), callable);
        }
    }

    fn process(&mut self, _delta: f64) {
        let object = self.get_target_object();
        let property = self.get_target_property();

        if let Some(mut target) = object {
            let value = map_range_clamped(
                self.pre_value,
                self.clamp_in_min,
                self.clamp_in_max,
                self.clamp_out_min,
                self.clamp_out_max,
            );
            target.set(property.clone(), Variant::from(value));
        }
    }
}
