use std::borrow::BorrowMut;

use godot::prelude::*;

use crate::{
    common::{map_blend_shape, map_range_clamped, ReadBlendShape},
    face_data_object::FaceDataObject,
    live_link_server::LiveLinkServer,
};

#[derive(GodotConvert, Var, Export, Debug)]
#[godot(via = GString)]
pub enum TransformAxis {
    X,
    Y,
    Z,
}

#[derive(GodotClass)]
#[class(base=Node)]
pub struct FaceData {
    #[export]
    read_from: Option<Gd<LiveLinkServer>>,
    #[export]
    read_shape: ReadBlendShape,

    #[export]
    target: Option<Gd<Node3D>>,

    #[export]
    clamp_in_min: f32,
    #[export]
    clamp_in_max: f32,
    #[export]
    clamp_out_min: f32,
    #[export]
    clamp_out_max: f32,

    #[export]
    transform_axis: TransformAxis,

    pre_value: f32,

    base: Base<Node>,
}

#[godot_api]
impl FaceData {
    #[func]
    fn on_face_data(&mut self, changed: bool, face_data: Option<Gd<FaceDataObject>>) {
        if changed {
            let face_data = face_data.unwrap();
            let shape_val = map_blend_shape(&self.read_shape, face_data);
            self.pre_value = shape_val;
        }
    }
}

#[godot_api]
impl INode for FaceData {
    fn init(base: Base<Node>) -> Self {
        Self {
            read_from: None,
            read_shape: ReadBlendShape::EyeBlinkLeft,
            target: None,
            clamp_in_min: 0.0,
            clamp_in_max: 1.0,
            clamp_out_min: 0.0,
            clamp_out_max: 1.0,

            transform_axis: TransformAxis::X,

            pre_value: 0.0,

            base,
        }
    }

    fn ready(&mut self) {
        if let Some(mut server) = self.borrow_mut().read_from.take() {
            let callable = self.base_mut().callable("on_face_data");
            server.connect("face_data_updated".into(), callable);
        }
    }

    fn process(&mut self, _delta: f64) {
        if let Some(target) = &mut self.target {
            let mut rotation = target.get_rotation_degrees();
            let clamped = map_range_clamped(
                self.pre_value,
                self.clamp_in_min,
                self.clamp_in_max,
                self.clamp_out_min,
                self.clamp_out_max,
            );

            match self.transform_axis {
                TransformAxis::X => {
                    rotation.x = clamped;
                }
                TransformAxis::Y => {
                    rotation.y = clamped;
                }
                TransformAxis::Z => {
                    rotation.z = clamped;
                }
            }

            target.set_rotation_degrees(rotation);
        }
    }
}
