use std::borrow::BorrowMut;

use godot::engine::global::Error;
use godot::engine::{Node, UdpServer};
use godot::obj::WithBaseField;
use godot::prelude::*;
use live_link_face::LiveLinkFace;

use crate::face_data_object::{read_face_data, FaceDataObject};

#[derive(GodotClass)]
#[class(base=Node)]
pub struct LiveLinkServer {
    #[export]
    listen_ip: GString,
    #[export]
    listen_port: u16,
    udp_server: Gd<UdpServer>,

    pub face_data: Option<LiveLinkFace>,

    base: Base<Node>,
}

#[godot_api]
impl LiveLinkServer {
    #[signal]
    fn face_data_updated(changed: bool, fd: Option<Gd<FaceDataObject>>);
}

#[godot_api]
impl INode for LiveLinkServer {
    fn init(base: Base<Node>) -> Self {
        let server = UdpServer::new_gd();

        Self {
            listen_ip: "100.64.0.2".into(),
            listen_port: 3011,
            udp_server: server,
            face_data: None,
            base,
        }
    }

    fn ready(&mut self) {
        let err = self
            .udp_server
            .listen_ex(self.listen_port)
            .bind_address(self.listen_ip.clone())
            .done();

        match err {
            Error::OK => {}
            _ => godot_error!("UdpServer listen error"),
        };
    }

    fn enter_tree(&mut self) {}

    fn process(&mut self, _delta: f64) {
        self.udp_server.poll();

        if self.udp_server.is_connection_available() {
            if let Some(mut peer) = self.udp_server.take_connection() {
                let packet = peer.get_packet();

                let (has_data, face_data) = LiveLinkFace::decode(packet.as_slice());
                if has_data {
                    self.borrow_mut().face_data = Some(face_data);
                    let mut face_data_object = FaceDataObject::new_gd();
                    read_face_data(
                        &mut face_data_object,
                        &self.borrow_mut().face_data.take().unwrap(),
                    );
                    let fdo_variant = Variant::from(Some(face_data_object));
                    let changed = Variant::from(has_data);
                    self.base_mut()
                        .emit_signal("face_data_updated".into(), &[changed, fdo_variant]);
                } else {
                    self.borrow_mut().face_data = None;
                    let changed = Variant::from(has_data);
                    let fdo_variant = Variant::from::<Option<Gd<FaceDataObject>>>(None);
                    self.base_mut()
                        .emit_signal("face_data_updated".into(), &[changed, fdo_variant]);
                }
            }
        }
    }

    fn exit_tree(&mut self) {}
}
